@@ -1,8 +1,7 @@
-import copy
 import inspect
 import logging
 from enum import Enum
-from typing import Type, Dict, Any, Callable, List
+from typing import Type, Dict, Any, Callable, List, Tuple, Union
 
 logger = logging.getLogger(__name__)
 
@@ -14,10 +13,14 @@ These get their own file because we don't like circular dependencies.
 
 
 class NodeSource(Enum):
-    """Specifies where this node's value originates. This can be used by different executors to flexibly execute a function graph."""
+    """
+    Specifies where this node's value originates.
+    This can be used by different adapters to flexibly execute a function graph.
+    """
     STANDARD = 1  # standard dependencies
     EXTERNAL = 2  # This node's value should be taken from cache
-    PRIOR_RUN = 3  # This node's value sould be taken from a prior run. This is not used in a standard function graph, but it comes in handy for repeatedly running the same one.
+    PRIOR_RUN = 3  # This node's value should be taken from a prior run.
+    # This is not used in a standard function graph, but it comes in handy for repeatedly running the same one.
 
 
 class DependencyType(Enum):
@@ -40,7 +43,7 @@ class Node(object):
                  doc_string: str = '',
                  callabl: Callable = None,
                  node_source: NodeSource = NodeSource.STANDARD,
-                 input_types: Dict[str, Type] = None):
+                 input_types: Dict[str, Union[Type, Tuple[Type, DependencyType]]] = None):
         """Constructor for our Node object.
 
         :param name: the name of the function.
@@ -60,23 +63,32 @@ class Node(object):
         self._dependencies = []
         self._depended_on_by = []
 
+        self._input_types = {}
+
         if self._node_source == NodeSource.STANDARD:
             if input_types is not None:
-                self._input_types = {key: (value, DependencyType.REQUIRED) for key, value in input_types.items()}
+                for key, value in input_types.items():
+                    if isinstance(value, tuple):
+                        self._input_types[key] = value
+                    else:
+                        self._input_types = {key: (value, DependencyType.REQUIRED) for key, value in input_types.items()}
             else:
                 signature = inspect.signature(callabl)
-                self._input_types = {}
                 for key, value in signature.parameters.items():
                     if value.annotation == inspect._empty:
                         raise ValueError(f'Missing type hint for {key} in function {name}. Please add one to fix.')
                     self._input_types[key] = (value.annotation, DependencyType.from_parameter(value))
+        elif self.user_defined:
+            if input_types is not None:
+                raise ValueError(f'Input types cannot be provided for user-defined node {self.name}')
+
 
     @property
     def documentation(self) -> str:
         return self._doc
 
     @property
-    def input_types(self) -> Dict[str, Type]:
+    def input_types(self) -> Dict[Any, Tuple[Any, DependencyType]]:
         return self._input_types
 
     @property
